"""Tests delivery report models validators."""

from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
from _pytest.logging import LogCaptureFixture
from pydantic import ValidationInfo

from cg.constants import NA_FIELD, NO_FIELD, REPORT_GENDER, YES_FIELD, Workflow
from cg.constants.constants import AnalysisType
from cg.constants.subject import Sex
from cg.models.delivery.delivery import DeliveryFile
from cg.models.orders.constants import OrderType
from cg.models.report.validators import (
    get_analysis_type_as_string,
    get_boolean_as_string,
    get_date_as_string,
    get_delivered_files_as_file_names,
    get_float_as_percentage,
    get_float_as_string,
    get_gender_as_string,
    get_list_as_string,
    get_path_as_string,
    get_prep_category_as_string,
    get_report_string,
)


def test_get_report_string():
    """Test formatting an empty value."""

    # GIVEN a not valid empty field
    none_field: Any = None

    # WHEN performing the validation
    output: str = get_report_string(none_field)

    # THEN check if the input value was formatted correctly
    assert output == NA_FIELD


def test_get_boolean_as_string():
    """Test boolean formatting for the delivery report."""

    # GIVEN a not formatted inputs
    none_field: Any = None
    true_field: bool = True
    false_field: bool = False
    not_bool_field: str = "not a boolean"

    # WHEN performing the validation
    validated_none_field: str = get_boolean_as_string(none_field)
    validated_true_field: str = get_boolean_as_string(true_field)
    validated_false_field: str = get_boolean_as_string(false_field)
    validated_not_bool_field = get_boolean_as_string(not_bool_field)

    # THEN check if the input values were formatted correctly
    assert validated_none_field == NA_FIELD
    assert validated_true_field == YES_FIELD
    assert validated_false_field == NO_FIELD
    assert validated_not_bool_field == NA_FIELD


def test_get_float_as_string():
    """Test the validation of a float value."""

    # GIVEN a valid float input
    float_value: float = 12.3456789

    # WHEN performing the validation
    validated_float_value: str = get_float_as_string(float_value)

    # THEN check if the input value was formatted correctly
    assert validated_float_value == "12.35"


def test_get_float_as_string_zero_input():
    """Tests the validation of a float value when input is zero."""

    # GIVEN a valid float input
    float_value: float = 0.0

    # WHEN performing the validation
    validated_float_value: str = get_float_as_string(float_value)

    # THEN check if the input value was formatted correctly
    assert validated_float_value == "0.0"


def test_get_float_as_percentage():
    """Test the validation of a percentage value."""

    # GIVEN a not formatted percentage
    pct_value: float = 0.9876

    # WHEN performing the validation
    validated_pct_value: str = get_float_as_percentage(pct_value)

    # THEN check if the input value was formatted correctly
    assert validated_pct_value == "98.76"


def test_get_float_as_percentage_zero_input():
    """Test the validation of a percentage value when input is zero."""

    # GIVEN a zero input
    pct_value: float = 0.0

    # WHEN performing the validation
    validated_pct_value: str = get_float_as_percentage(pct_value)

    # THEN check if the input value was formatted correctly
    assert validated_pct_value == "0.0"


def test_get_date_as_string(timestamp_now: datetime):
    """Test the validation of a datetime object."""

    # GIVEN a datetime object

    # WHEN performing the validation
    validated_date_value: str = get_date_as_string(timestamp_now)

    # THEN check if the input values were formatted correctly
    assert validated_date_value == timestamp_now.date().__str__()


def test_get_list_as_string():
    """Test if a list is transformed into a string of comma separated values."""

    # GIVEN a mock list
    mock_list: list[str] = ["I am", "a", "list"]

    # WHEN performing the validation
    validated_list: str = get_list_as_string(mock_list)

    # THEN check if the input values were formatted correctly
    assert validated_list == "I am, a, list"


def test_get_list_paths_as_strings(filled_file: Path):
    """Test file path name extraction from a list."""

    # GIVEN a list of delivery files
    path_list: list[DeliveryFile] = [
        DeliveryFile(source_path=filled_file, destination_path=filled_file)
    ]

    # WHEN validating the provided list of delivery paths
    path_name_list: list[str] = get_delivered_files_as_file_names(path_list)

    # THEN the returned list should contain the file names
    assert path_name_list.pop() == "a_file.txt"


def test_get_path_as_string(filled_file: Path):
    """Test file path name extraction."""

    # GIVEN a mock path

    # WHEN performing the validation
    path_name: str = get_path_as_string(filled_file)

    # THEN check if the input values were formatted correctly
    assert path_name == "a_file.txt"


def test_get_gender_as_string():
    """Test report gender parsing."""

    # GIVEN an invalid gender category
    gender: Sex = Sex.FEMALE
    invalid_gender: str = "not_a_gender"

    # WHEN performing the validation
    validated_gender: str = get_gender_as_string(gender)
    validated_invalid_gender: str = get_gender_as_string(invalid_gender)

    # THEN check if the gender has been correctly formatted
    assert validated_gender == REPORT_GENDER.get("female")
    assert validated_invalid_gender == NA_FIELD


def test_get_prep_category_as_string(caplog: LogCaptureFixture):
    """Test validation on a preparation category."""

    # GIVEN an invalid prep category
    prep_category: OrderType = OrderType.RML

    # WHEN performing the validation

    # THEN check if an exception was raised
    with pytest.raises(ValueError):
        get_prep_category_as_string(prep_category)
    assert "The delivery report generation does not support RML samples" in caplog.text


def test_get_analysis_type_as_string():
    """Test analysis type formatting for the delivery report generation."""

    # GIVEN a WGS analysis type and a model info dictionary
    analysis_type: AnalysisType = AnalysisType.WHOLE_GENOME_SEQUENCING
    model_info: ValidationInfo = ValidationInfo
    model_info.data: dict[str, Any] = {"workflow": Workflow.MIP_DNA.value}

    # WHEN performing the validation
    validated_analysis_type: str = get_analysis_type_as_string(
        analysis_type=analysis_type, info=model_info
    )

    # THEN check if the input value was formatted correctly
    assert validated_analysis_type == analysis_type.value


def test_get_analysis_type_as_string_balsamic():
    """Test analysis type formatting for the delivery report generation."""

    # GIVEN a WGS analysis type and a model info dictionary
    analysis_type: str = "tumor_normal_wgs"
    model_info: ValidationInfo = ValidationInfo
    model_info.data: dict[str, Any] = {"workflow": Workflow.BALSAMIC.value}

    # WHEN performing the validation
    validated_analysis_type: str = get_analysis_type_as_string(
        analysis_type=analysis_type, info=model_info
    )

    # THEN check if the input value was formatted correctly
    assert validated_analysis_type == "Tumör/normal (helgenomsekvensering)"



import click
from click.core import ParameterSource

from cg.constants.archiving import DEFAULT_SPRING_ARCHIVE_COUNT
from cg.constants.constants import DRY_RUN
from cg.meta.archive.archive import SpringArchiveAPI
from cg.models.cg_config import CGConfig


@click.group()
def archive():
    """Archive utilities."""
    pass


@archive.command("archive-spring-files")
@click.option(
    "-l",
    "--limit",
    help="Give a limit to the amount of files to archive.",
    default=DEFAULT_SPRING_ARCHIVE_COUNT,
    show_default=True,
)
@click.option(
    "--archive-all",
    help="Use in order to archive all non-archived-files",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_obj
def archive_spring_files(context: CGConfig, limit: int | None, archive_all: bool):
    """Archives non-archived spring files.
    Raises:
        click.Abort if both a limit to the number of spring files to archive and archive_all is specified.
    """

    if (
        click.get_current_context().get_parameter_source("limit") == ParameterSource.COMMANDLINE
        and limit
        and archive_all
    ):
        click.echo(
            "Incorrect input parameters - please do not provide both a limit and set --archive-all."
        )
        raise click.Abort

    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.archive_spring_files_and_add_archives_to_housekeeper(
        spring_file_count_limit=None if archive_all else limit
    )


@archive.command("update-job-statuses")
@click.pass_obj
def update_job_statuses(context: CGConfig):
    """Queries ongoing jobs and updates Housekeeper."""
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.update_statuses_for_ongoing_tasks()


@archive.command("delete-file")
@DRY_RUN
@click.pass_obj
@click.argument("file_path", required=True)
def delete_file(context: CGConfig, dry_run: bool, file_path: str):
    """Delete an archived file and remove it from Housekeeper.
    The file will not be deleted if it is not confirmed archived.
    The file will not be deleted if its archive location can not be determined from the file tags.
    """
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.delete_file(file_path=file_path, dry_run=dry_run)


@archive.group("retrieve")
def retrieve_spring():
    """Retrieve spring files."""
    pass


@retrieve_spring.command("sample")
@click.pass_obj
@click.argument("sample_id", required=True)
def retrieve_spring_files_for_sample(context: CGConfig, sample_id: str):
    """Retrieve spring files for a sample."""
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.retrieve_spring_files_for_sample(sample_id=sample_id)


@retrieve_spring.command("case")
@click.pass_obj
@click.argument("case_id", required=True)
def retrieve_spring_files_for_case(context: CGConfig, case_id: str):
    """Retrieve spring files for a case."""
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.retrieve_spring_files_for_case(case_id=case_id)


@retrieve_spring.command("order")
@click.pass_obj
@click.option("-t", "--ticket_id", help="Specify ticket_id instead of order_id")
@click.argument("order_id", required=False)
def retrieve_spring_files_for_order(context: CGConfig, order_id: int = None, ticket_id: int = None):
    """Retrieve spring files for an order."""
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow,
    )
    spring_archive_api.retrieve_spring_files_for_order(
        id_=order_id or ticket_id, is_order_id=order_id is not None
    )



import logging
from collections import defaultdict

from pydantic import BaseModel

from cg.apps.demultiplex.sample_sheet.sample_models import FlowCellSample

LOG = logging.getLogger(__name__)


class SampleSheet(BaseModel):
    samples: list[FlowCellSample]

    def get_non_pooled_lanes_and_samples(self) -> list[tuple[int, str]]:
        """Return tuples of non-pooled lane and sample ids."""
        non_pooled_lane_sample_id_pairs: list[tuple[int, str]] = []
        non_pooled_samples: list[FlowCellSample] = self.get_non_pooled_samples()
        for sample in non_pooled_samples:
            non_pooled_lane_sample_id_pairs.append((sample.lane, sample.sample_id))
        return non_pooled_lane_sample_id_pairs

    def get_non_pooled_samples(self) -> list[FlowCellSample]:
        """Return samples that are sequenced solo in their lane."""
        lane_samples: dict[int, list[FlowCellSample]] = defaultdict(list)
        for sample in self.samples:
            lane_samples[sample.lane].append(sample)
        return [samples[0] for samples in lane_samples.values() if len(samples) == 1]

    def get_sample_ids(self) -> list[str]:
        """Return ids for samples in sheet."""
        sample_internal_ids: list[str] = []
        for sample in self.samples:
            sample_internal_ids.append(sample.sample_id)
        return list(set(sample_internal_ids))



import logging
from pathlib import Path

import click
from pydantic import ValidationError

from cg.apps.demultiplex.sample_sheet.api import SampleSheetAPI
from cg.constants.constants import DRY_RUN
from cg.exc import SampleSheetError
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(name="samplesheet")
def sample_sheet_commands():
    """Command group for the sample sheet commands."""


@sample_sheet_commands.command(name="validate")
@click.argument("sheet", type=click.Path(exists=True, dir_okay=False))
@click.pass_obj
def validate_sample_sheet(context: CGConfig, sheet: click.Path):
    """
    Validate a sample sheet.
    """
    LOG.info(f"Validating {sheet} sample sheet")
    sample_sheet_api: SampleSheetAPI = context.sample_sheet_api
    try:
        sample_sheet_api.validate_sample_sheet(Path(sheet))
    except (SampleSheetError, ValidationError) as error:
        LOG.error("Sample sheet failed validation")
        raise click.Abort from error
    LOG.info("Sample sheet passed validation")


@sample_sheet_commands.command(name="translate")
@click.argument("flow-cell-name")
@DRY_RUN
@click.pass_obj
def translate_sample_sheet(context: CGConfig, flow_cell_name: str, dry_run: bool):
    """
    Translate a sample sheet from Bcl2Fastq to BclConvert format.
    'flow-cell-name' is the flow cell run directory name, e.g. '181005_D00410_0735_BHM2LNBCX2' with
    a Bcl2Fastq sample sheet in it.
    """
    LOG.info(f"Translating Bcl2Fastq sample sheet for flow cell {flow_cell_name}")
    sample_sheet_api: SampleSheetAPI = context.sample_sheet_api
    sample_sheet_api.set_dry_run(dry_run)
    sample_sheet_api.translate_sample_sheet(flow_cell_name=flow_cell_name)


@sample_sheet_commands.command(name="create")
@click.argument("flow-cell-name")
@DRY_RUN
@click.option("--force", is_flag=True, help="Skips the validation of the sample sheet")
@click.pass_obj
def create_sheet(
    context: CGConfig,
    flow_cell_name: str,
    dry_run: bool,
    force: bool = False,
):
    """Create a sample sheet or hard-link it from Housekeeper in the flow cell directory.

    'flow-cell-name' is the flow cell run directory name, e.g. '181005_D00410_0735_BHM2LNBCX2'

    Search the flow cell in the directory specified in config.
    """
    LOG.info(f"Getting a valid sample sheet for flow cell {flow_cell_name}")
    sample_sheet_api: SampleSheetAPI = context.sample_sheet_api
    sample_sheet_api.set_dry_run(dry_run)
    sample_sheet_api.set_force(force)
    sample_sheet_api.get_or_create_sample_sheet(flow_cell_name)


@sample_sheet_commands.command(name="create-all")
@DRY_RUN
@click.pass_obj
def create_all_sheets(context: CGConfig, dry_run: bool):
    """Command to create sample sheets for all flow cells that lack a sample sheet.

    Search flow cell directories for run parameters and create a sample sheets based on the
    information.
    """
    sample_sheet_api: SampleSheetAPI = context.sample_sheet_api
    sample_sheet_api.set_dry_run(dry_run)
    sample_sheet_api.set_force(force=False)
    sample_sheet_api.get_or_create_all_sample_sheets()



from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.services.orders.order_summary_service.dto.case_summary import CaseSummary
from cg.services.orders.order_summary_service.dto.order_summary import OrderSummary
from cg.services.orders.order_summary_service.utils import (
    create_summaries,
    get_cases_failed_sequencing_qc_count,
)
from cg.store.models import Order
from cg.store.store import Store


class OrderSummaryService:
    def __init__(self, analysis_client: TrailblazerAPI, store: Store) -> None:
        self.analysis_client = analysis_client
        self.store = store

    def get_summary(self, order_id: int) -> OrderSummary:
        return self.get_summaries([order_id])[0]

    def get_summaries(self, order_ids: list[int]) -> list[OrderSummary]:
        orders: list[Order] = self.store.get_orders_by_ids(order_ids)
        cases: list[CaseSummary] = self._get_case_summaries(order_ids)
        analyses: list[AnalysisSummary] = self.analysis_client.get_summaries(order_ids)
        return create_summaries(orders=orders, analysis_summaries=analyses, case_summaries=cases)

    def _get_case_summaries(self, order_ids: list[int]):
        summaries: list[CaseSummary] = []

        for order_id in order_ids:
            not_received: int = self.store.get_case_not_received_count(order_id)
            in_preparation: int = self.store.get_case_in_preparation_count(order_id)
            in_sequencing: int = self.store.get_case_in_sequencing_count(order_id)
            order: Order = self.store.get_order_by_id(order_id)
            failed_sequencing_qc: int = get_cases_failed_sequencing_qc_count(order)

            summary = CaseSummary(
                order_id=order_id,
                not_received=not_received,
                in_preparation=in_preparation,
                in_sequencing=in_sequencing,
                failed_sequencing_qc=failed_sequencing_qc,
            )
            summaries.append(summary)

        return summaries



"""Helper functions for observations related actions."""

import logging

from sqlalchemy.orm import Query

from cg.constants.constants import Workflow
from cg.constants.observations import LOQUSDB_SUPPORTED_WORKFLOWS
from cg.constants.sequencing import SequencingMethod
from cg.exc import CaseNotFoundError, LoqusdbUploadCaseError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


def get_observations_case(context: CGConfig, case_id: str, upload: bool) -> Case:
    """Return a verified Loqusdb case."""
    status_db: Store = context.status_db
    case: Case = status_db.get_case_by_internal_id(internal_id=case_id)
    if not case or case.data_analysis not in LOQUSDB_SUPPORTED_WORKFLOWS:
        LOG.error("Invalid case ID. Retrieving available cases for Loqusdb actions.")
        cases_to_process: Query = (
            status_db.observations_to_upload() if upload else status_db.observations_uploaded()
        )
        if not cases_to_process:
            LOG.info("There are no valid cases to be processed by Loqusdb")
        else:
            LOG.info("Provide one of the following case IDs: ")
            for case in cases_to_process:
                LOG.info(f"{case.internal_id} ({case.data_analysis})")

        raise CaseNotFoundError
    return case


def get_observations_case_to_upload(context: CGConfig, case_id: str) -> Case:
    """Return a verified case ready to be uploaded to Loqusdb."""
    case: Case = get_observations_case(context, case_id, upload=True)
    if not case.customer.loqus_upload:
        LOG.error(
            f"Customer {case.customer.internal_id} is not whitelisted for upload to Loqusdb. Canceling upload for "
            f"case {case.internal_id}."
        )
        raise LoqusdbUploadCaseError
    return case


def get_observations_api(
    context: CGConfig, case: Case
) -> MipDNAObservationsAPI | BalsamicObservationsAPI:
    """Return an observations API given a specific case object."""
    observations_apis = {
        Workflow.MIP_DNA: MipDNAObservationsAPI(context, get_sequencing_method(case)),
        Workflow.BALSAMIC: BalsamicObservationsAPI(context, get_sequencing_method(case)),
    }
    return observations_apis[case.data_analysis]


def get_sequencing_method(case: Case) -> SequencingMethod:
    """Returns the sequencing method for the given case object."""
    analysis_types = [
        link.sample.application_version.application.analysis_type for link in case.links
    ]
    if len(set(analysis_types)) != 1:
        LOG.error(f"Case {case.internal_id} has a mixed analysis type. Cancelling action.")
        raise LoqusdbUploadCaseError

    return analysis_types[0]



"""Module for Raredisease Analysis API."""

import logging
from pathlib import Path
from typing import Any

from cg.constants import Workflow
from cg.constants.constants import GenomeVersion
from cg.constants.gene_panel import GenePanelGenomeBuild
from cg.constants.nf_analysis import RAREDISEASE_METRIC_CONDITIONS
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.nf_analysis import WorkflowParameters
from cg.models.raredisease.raredisease import (
    RarediseaseSampleSheetEntry,
    RarediseaseSampleSheetHeaders,
)
from cg.store.models import CaseSample, Sample

LOG = logging.getLogger(__name__)


class RarediseaseAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RAREDISEASE processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RAREDISEASE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.raredisease.root
        self.nfcore_workflow_path: str = config.raredisease.workflow_path
        self.references: str = config.raredisease.references
        self.profile: str = config.raredisease.profile
        self.conda_env: str = config.raredisease.conda_env
        self.conda_binary: str = config.raredisease.conda_binary
        self.config_platform: str = config.raredisease.config_platform
        self.config_params: str = config.raredisease.config_params
        self.config_resources: str = config.raredisease.config_resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.raredisease.tower_workflow
        self.account: str = config.raredisease.slurm.account
        self.email: str = config.raredisease.slurm.mail_user
        self.compute_env_base: str = config.raredisease.compute_env
        self.revision: str = config.raredisease.revision
        self.nextflow_binary_path: str = config.raredisease.binary_path

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return RarediseaseSampleSheetHeaders.list()

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a case. Currently fixed for hg19."""
        return GenomeVersion.hg19

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        fastq_forward_read_paths, fastq_reverse_read_paths = self.get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = RarediseaseSampleSheetEntry(
            name=case_sample.sample.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            sex=self.get_sex_code(case_sample.sample.sex),
            phenotype=self.get_phenotype_code(case_sample.status),
            paternal_id=case_sample.get_paternal_sample_id,
            maternal_id=case_sample.get_maternal_sample_id,
            case_id=case_sample.case.internal_id,
        )
        return sample_sheet_entry.reformat_sample_content

    def get_workflow_parameters(self, case_id: str) -> WorkflowParameters:
        """Return parameters."""
        return WorkflowParameters(
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
        )

    @staticmethod
    def get_phenotype_code(phenotype: str) -> int:
        """Return Raredisease phenotype code."""
        LOG.debug("Translate phenotype to integer code")
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def get_sex_code(sex: str) -> int:
        """Return Raredisease sex code."""
        LOG.debug("Translate sex to integer code")
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code

    @property
    def root(self) -> str:
        return self.config.raredisease.root

    def write_managed_variants(self, case_id: str, content: list[str]) -> None:
        self._write_managed_variants(out_dir=Path(self.root, case_id), content=content)

    def get_managed_variants(self) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(genome_build=GenePanelGenomeBuild.hg19)

    def get_workflow_metrics(self, sample_id: str) -> dict:
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        metric_conditions: dict[str, dict[str, Any]] = dict(RAREDISEASE_METRIC_CONDITIONS)
        self.set_order_sex_for_sample(sample, metric_conditions)
        return metric_conditions

    @staticmethod
    def set_order_sex_for_sample(sample: Sample, metric_conditions: dict) -> None:
        metric_conditions["predicted_sex_sex_check"]["threshold"] = sample.sex



"""Functions to get sample sheet information from Lims."""

import logging
import re
from typing import Iterable, Type

from genologics.entities import Artifact, Container, Sample
from genologics.lims import Lims

from cg.apps.demultiplex.sample_sheet.sample_models import FlowCellSample

LOG = logging.getLogger(__name__)


def get_placement_lane(lane: str) -> int:
    """Parse out the lane information from an artifact.placement."""
    return int(lane.split(":")[0])


def get_non_pooled_artifacts(artifact: Artifact) -> list[Artifact]:
    """Find the parent artifact of the sample. Should hold the reagent_label."""
    artifacts = []

    if len(artifact.samples) == 1:
        artifacts.append(artifact)
        return artifacts

    for artifact_input in artifact.input_artifact_list():
        artifacts.extend(get_non_pooled_artifacts(artifact_input))

    return artifacts


def get_reagent_label(artifact) -> str | None:
    """Get the first and only reagent label from an artifact."""
    labels: list[str] = artifact.reagent_labels
    if len(labels) > 1:
        raise ValueError(f"Expecting at most one reagent label. Got ({len(labels)}).")
    return labels[0] if labels else None


def extract_sequence_in_parentheses(label: str) -> str | None:
    """Return the sequence in parentheses from the reagent label or None if not found."""
    match = re.match(r"^[^(]+ \(([^)]+)\)$", label)
    return match.group(1) if match else None


def get_index(lims: Lims, label: str) -> str:
    """Parse out the sequence from a reagent label."""

    reagent_types = lims.get_reagent_types(name=label)

    if len(reagent_types) > 1:
        raise ValueError(f"Expecting at most one reagent type. Got ({len(reagent_types)}).")

    try:
        reagent_type = reagent_types.pop()
    except IndexError:
        return ""
    sequence: str = reagent_type.sequence

    match = extract_sequence_in_parentheses(label=label)
    if match:
        assert match == sequence

    return sequence


def get_flow_cell_samples(
    lims: Lims,
    flow_cell_id: str,
) -> Iterable[FlowCellSample]:
    """Return samples from LIMS for a given flow cell."""
    LOG.info(f"Fetching samples from lims for flowcell {flow_cell_id}")
    containers: list[Container] = lims.get_containers(name=flow_cell_id)
    if not containers:
        return []
    container: Container = containers[-1]  # only take the last one. See ÖA#217.
    raw_lanes: list[str] = sorted(container.placements.keys())
    for raw_lane in raw_lanes:
        lane: int = get_placement_lane(raw_lane)
        placement_artifact: Artifact = container.placements[raw_lane]
        non_pooled_artifacts: list[Artifact] = get_non_pooled_artifacts(placement_artifact)
        for artifact in non_pooled_artifacts:
            sample: Sample = artifact.samples[0]  # we are assured it only has one sample
            label: str | None = get_reagent_label(artifact)
            index = get_index(lims=lims, label=label)
            yield FlowCellSample(
                lane=lane,
                sample_id=sample.id,
                index=index,
            )



import math

from cg.meta.demultiplex.combine_sequencing_metrics import (
    combine_mapped_metrics_with_undetermined,
    combine_metrics,
    weighted_average,
)
from cg.store.models import SampleLaneSequencingMetrics


def test_calculates_simple_weighted_average():
    # GIVEN Equal total counts and different percentages
    total_1, percentage_1 = 50, 0.9
    total_2, percentage_2 = 50, 0.7

    # WHEN Calculating the weighted average
    result: float = weighted_average(
        total_1=total_1, percentage_1=percentage_1, total_2=total_2, percentage_2=percentage_2
    )

    # THEN The weighted average should be 0.8
    assert math.isclose(result, 0.8, rel_tol=1e-9)


def test_handles_zero_counts():
    # GIVEN zero counts for totals
    total_1, percentage_1 = 0, 0.0
    total_2, percentage_2 = 0, 0.0

    # WHEN Calculating the weighted average
    result: float = weighted_average(
        total_1=total_1, percentage_1=percentage_1, total_2=total_2, percentage_2=percentage_2
    )

    # THEN The weighted average should be zero
    assert result == 0


def test_combine_metrics():
    # GIVEN two metrics
    existing_metric = SampleLaneSequencingMetrics(
        sample_total_reads_in_lane=100,
        sample_base_percentage_passing_q30=0.9,
        sample_base_mean_quality_score=30,
    )
    new_metric = SampleLaneSequencingMetrics(
        sample_total_reads_in_lane=100,
        sample_base_percentage_passing_q30=0.8,
        sample_base_mean_quality_score=25,
    )

    # WHEN Combining the metrics
    combine_metrics(existing_metric=existing_metric, new_metric=new_metric)

    # THEN The existing metric should be updated
    assert existing_metric.sample_total_reads_in_lane == 200
    assert math.isclose(existing_metric.sample_base_percentage_passing_q30, 0.85, rel_tol=1e-9)
    assert math.isclose(existing_metric.sample_base_mean_quality_score, 27.5, rel_tol=1e-9)


def test_combine_empty_metrics():
    # GIVEN empty lists for mapped and undetermined metrics
    mapped_metrics = []
    undetermined_metrics = []

    # WHEN combining them
    combined_metrics = combine_mapped_metrics_with_undetermined(
        mapped_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be an empty list
    assert combined_metrics == []


def test_combine_metrics_with_only_mapped_metrics():
    # GIVEN one mapped metric and no undetermined
    mapped_metrics = [SampleLaneSequencingMetrics()]
    undetermined_metrics = []

    # WHEN combining them
    combined_metrics = combine_mapped_metrics_with_undetermined(
        mapped_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be the mapped metrics
    assert combined_metrics == mapped_metrics


def test_combine_metrics_with_only_undetermined_metrics():
    # GIVEN an empty list of mapped metrics and list of undetermined metrics
    mapped_metrics = []
    undetermined_metrics = [SampleLaneSequencingMetrics()]

    # WHEN combining them
    combined_metrics = combine_mapped_metrics_with_undetermined(
        mapped_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be the undetermined metrics
    assert combined_metrics == undetermined_metrics


def test_combine_metrics_with_both_mapped_and_undetermined_metrics_different_lanes():
    # GIVEN one mapped and one undetermined metric in different lanes for a sample
    mapped_metrics = [
        SampleLaneSequencingMetrics(flow_cell_lane_number=1, sample_internal_id="sample")
    ]
    undetermined_metrics = [
        SampleLaneSequencingMetrics(flow_cell_lane_number=2, sample_internal_id="sample")
    ]

    # WHEN combining them
    combined_metrics = combine_mapped_metrics_with_undetermined(
        mapped_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN two metrics should be returned
    assert len(combined_metrics) == 2
    assert set(combined_metrics) == set(mapped_metrics + undetermined_metrics)


def test_combine_metrics_with_both_mapped_and_undetermined_metrics_same_lane():
    # GIVEN a list of mapped metrics and list of undetermined metrics in the same lane for the same sample
    mapped_metric = SampleLaneSequencingMetrics(
        sample_internal_id="sample",
        flow_cell_lane_number=1,
        sample_total_reads_in_lane=100,
        sample_base_percentage_passing_q30=0.9,
        sample_base_mean_quality_score=30,
    )

    undetermined_metric = SampleLaneSequencingMetrics(
        sample_internal_id="sample",
        flow_cell_lane_number=1,
        sample_total_reads_in_lane=100,
        sample_base_percentage_passing_q30=0.8,
        sample_base_mean_quality_score=20,
    )

    # WHEN combining them
    combined_metrics = combine_mapped_metrics_with_undetermined(
        mapped_metrics=[mapped_metric], undetermined_metrics=[undetermined_metric]
    )

    # THEN the combined metrics should be a single metric
    assert len(combined_metrics) == 1

    # THEN the metrics should be a weighted average of the mapped and undetermined metrics
    metric: SampleLaneSequencingMetrics = combined_metrics[0]
    assert metric.sample_total_reads_in_lane == 200
    assert math.isclose(metric.sample_base_percentage_passing_q30, 0.85, rel_tol=1e-9)
    assert metric.sample_base_mean_quality_score == 25



import logging
from pathlib import Path

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import FlowCellError, MissingFilesError
from cg.meta.demultiplex.utils import get_sample_fastqs_from_flow_cell
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)


def is_demultiplexing_complete(flow_cell_directory: Path) -> bool:
    return Path(flow_cell_directory, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def is_flow_cell_ready_for_delivery(flow_cell_directory: Path) -> bool:
    return Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).exists()


def validate_sample_sheet_exists(flow_cell: FlowCellDirectoryData) -> None:
    sample_sheet_path: Path = flow_cell.get_sample_sheet_path_hk()
    if not sample_sheet_path or not sample_sheet_path.exists():
        raise FlowCellError(f"Sample sheet {sample_sheet_path} does not exist in housekeeper.")
    LOG.debug(f"Found sample sheet {sample_sheet_path} in housekeeper.")


def validate_demultiplexing_complete(flow_cell_output_directory: Path) -> None:
    if not is_demultiplexing_complete(flow_cell_output_directory):
        raise FlowCellError(
            f"Demultiplexing not completed for flow cell directory {flow_cell_output_directory}."
        )


def validate_flow_cell_delivery_status(flow_cell_output_directory: Path, force: bool) -> None:
    if is_flow_cell_ready_for_delivery(flow_cell_output_directory) and not force:
        raise FlowCellError(
            f"Flow cell output directory {flow_cell_output_directory}"
            " has already been processed and is ready for delivery."
        )


def validate_flow_cell_has_fastq_files(flow_cell: FlowCellDirectoryData) -> None:
    """Check if any sample from the flow cell has fastq files.
    Raises: MissingFilesError
        When all samples have missing fastq files in the flow cell
    """
    sample_ids: list[str] = flow_cell.sample_sheet.get_sample_ids()
    for sample_id in sample_ids:
        fastq_files: list[Path] | None = get_sample_fastqs_from_flow_cell(
            flow_cell_directory=flow_cell.path, sample_internal_id=sample_id
        )
        if fastq_files:
            LOG.debug(f"Flow cell {flow_cell.id} has at least one sample with fastq files")
            return
    raise MissingFilesError(f"No fastq files were found for any sample in flow cell {flow_cell.id}")


def is_flow_cell_ready_for_postprocessing(
    flow_cell_output_directory: Path,
    flow_cell: FlowCellDirectoryData,
    force: bool = False,
) -> None:
    validate_flow_cell_delivery_status(
        flow_cell_output_directory=flow_cell_output_directory, force=force
    )
    validate_sample_sheet_exists(flow_cell)
    validate_demultiplexing_complete(flow_cell_output_directory)
    validate_flow_cell_has_fastq_files(flow_cell)



from datetime import datetime
from pathlib import Path
from typing import Any

import pytest

from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections, SampleSheetBCLConvertSections
from cg.io.controller import ReadFile


@pytest.fixture(scope="function")
def sample_sheet_validator() -> SampleSheetValidator:
    """Return a sample sheet validator."""
    return SampleSheetValidator()


@pytest.fixture
def hiseq_x_single_index_sample_sheet_content(
    hiseq_x_single_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a single-index HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_single_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_x_dual_index_sample_sheet_content(
    hiseq_x_dual_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a dual-index HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_dual_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_dual_index_sample_sheet_content(
    hiseq_2500_dual_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a dual-index HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_dual_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_custom_index_sample_sheet_content(
    hiseq_2500_custom_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a custom-index HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_custom_index_sample_sheet_path
    )


@pytest.fixture
def novaseq_6000_pre_1_5_kits_sample_sheet_content(
    novaseq_6000_pre_1_5_kits_correct_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_pre_1_5_kits_correct_sample_sheet_path
    )


@pytest.fixture
def novaseq_6000_post_1_5_kits_sample_sheet_content(
    novaseq_6000_post_1_5_kits_correct_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_post_1_5_kits_correct_sample_sheet_path
    )


@pytest.fixture
def novaseq_x_sample_sheet_content(novaseq_x_correct_sample_sheet: Path) -> list[list[str]]:
    """Return the content of a NovaSeqX sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_x_correct_sample_sheet
    )


# Sample sheet parts


@pytest.fixture
def bcl_convert_sample_sheet_line_entry_1() -> list[str]:
    """Return a sample sheet line entry from the HiSeqX double index flow cell."""
    return [
        "1",
        "ACC4519A1",
        "ATTCAGAA",
        "TATAGCCT",
        "Y151;I8;I8;Y151",
        "1",
        "1",
    ]


@pytest.fixture
def bcl_convert_sample_sheet_line_entry_2() -> list[str]:
    """Return a sample sheet line entry from the HiSeqX double index flow cell."""
    return [
        "2",
        "ACC4519A2",
        "GAATTCGT",
        "TATAGCCT",
        "Y151;I8;I8;Y151",
        "1",
        "1",
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header() -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture
def sample_sheet_bcl_convert_data_header() -> list[list[str]]:
    """Return the content of a BCLConvert sample sheet data header without samples."""
    return [
        [SampleSheetBCLConvertSections.Data.HEADER],
        [
            SampleSheetBCLConvertSections.Data.LANE.value,
            SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID.value,
            SampleSheetBCLConvertSections.Data.INDEX_1.value,
            SampleSheetBCLConvertSections.Data.INDEX_2.value,
            SampleSheetBCLConvertSections.Data.OVERRIDE_CYCLES.value,
            SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_1.value,
            SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_2.value,
        ],
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header_with_replaced_sample_id() -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


# Incorrect sample sheets


@pytest.fixture
def sample_sheet_content_missing_data_header() -> list[list[str]]:
    """Return a sample sheet without data and without the data header."""
    return [["[Header]"], ["[Reads]"], ["[BCLConvert_Settings]"]]


@pytest.fixture
def sample_sheet_samples_no_column_names(
    bcl_convert_sample_sheet_line_entry_1: list[str],
    bcl_convert_sample_sheet_line_entry_2: list[str],
) -> list[list[str]]:
    """Return the content of a sample sheet with samples but without the column names."""
    return [
        [SampleSheetBCLConvertSections.Data.HEADER],
        bcl_convert_sample_sheet_line_entry_1,
        bcl_convert_sample_sheet_line_entry_2,
    ]


@pytest.fixture
def novaseq_6000_sample_sheet_with_reversed_cycles_content(
    novaseq_6000_sample_sheet_with_reversed_cycles: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet with reversed cycles."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_sample_sheet_with_reversed_cycles
    )


@pytest.fixture
def novaseq_x_sample_sheet_with_forward_cycles_content(
    novaseq_x_sample_sheet_with_forward_cycles: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeqX sample sheet with forward cycles."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_x_sample_sheet_with_forward_cycles
    )


# Sample sheet objects


@pytest.fixture
def novaseq_6000_post_1_5_kits_sample_sheet_object(
    sample_sheet_validator: SampleSheetValidator,
    novaseq_6000_post_1_5_kits_correct_sample_sheet_path: Path,
) -> SampleSheet:
    """Return a NovaSeq 6000 sample sheet object."""
    return sample_sheet_validator.get_sample_sheet_object_from_file(
        file_path=novaseq_6000_post_1_5_kits_correct_sample_sheet_path
    )


# HK bundle objects


@pytest.fixture
def sample_sheet_bcl2fastq_bundle_data(
    tmp_flow_cell_with_bcl2fastq_sample_sheet: Path,
    timestamp_yesterday: datetime,
) -> dict[str, Any]:
    """Return a sample sheet bundle data dictionary."""
    flow_cell_id: str = tmp_flow_cell_with_bcl2fastq_sample_sheet.name[-9:]
    return {
        "name": flow_cell_id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [
            {
                "path": Path(
                    tmp_flow_cell_with_bcl2fastq_sample_sheet, "SampleSheet.csv"
                ).as_posix(),
                "archive": False,
                "tags": ["samplesheet", flow_cell_id],
            },
        ],
    }



"""Test CLI functions to get flow cells in the Status database."""

from datetime import datetime

from click.testing import CliRunner

from cg.cli.get import get
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store.models import Flowcell, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_get_flow_cell_bad_flow_cell(cli_runner: CliRunner, base_context: CGConfig):
    """Test that trying to get a flow cell using a non-existing flow cell fails."""
    # GIVEN an empty database

    # WHEN getting a flow cell
    result = cli_runner.invoke(get, ["flow-cell", "does_not_exist"], obj=base_context)

    # THEN it should complain about missing flow cell instead of getting a flow cell
    assert result.exit_code == EXIT_FAIL


def test_get_flow_cell_required(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test to get a flow cell using only the required arguments."""
    # GIVEN a database with a flow cell
    flow_cell: Flowcell = helpers.add_flow_cell(store=disk_store)

    # WHEN getting a flow cell
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS


def test_get_flow_cell_output(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flow cell with data
    flow_cell: Flowcell = helpers.add_flow_cell(disk_store)

    # WHEN getting a flow cell
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should have been returned
    assert flow_cell.name in result.output
    assert flow_cell.sequencer_type in result.output
    assert flow_cell.sequencer_name in result.output
    assert str(flow_cell.sequenced_at.date()) in result.output
    assert flow_cell.status in result.output


def test_get_flow_cell_archived_at_none(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output when archived at is none."""
    # GIVEN a database with a flow cell with data
    flow_cell: Flowcell = helpers.add_flow_cell(store=disk_store, archived_at=None)

    # WHEN getting a flow cell
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN log that the flow cell is on disk
    assert "ondisk" in result.output


def test_get_flow_cell_archived_at_date(
    cli_runner: CliRunner,
    base_context: CGConfig,
    disk_store: Store,
    helpers: StoreHelpers,
    timestamp: datetime,
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell with data
    flow_cell: Flowcell = helpers.add_flow_cell(store=disk_store, archived_at=timestamp)

    # WHEN getting a flow cell
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should have been returned
    assert str(timestamp.date()) in result.output


def test_get_flow_cell_samples_without_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers, caplog
):
    """Test that the output without samples."""
    # GIVEN a database with a flow cell without related samples
    flow_cell: Flowcell = helpers.add_flow_cell(disk_store)

    # WHEN getting a flow cell with the --samples flag
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name, "--samples"], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN a message about no samples should have been displayed
    assert "No samples found on flow cell" in caplog.text


def test_get_flow_cell_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the samples of the flow cell."""
    # GIVEN a database with a flow cell with related samples
    samples: list[Sample] = helpers.add_samples(store=disk_store)
    flow_cell: Flowcell = helpers.add_flow_cell(store=disk_store, samples=samples)

    # WHEN getting a flow cell with the --samples flag
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name, "--samples"], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN all related samples should be listed in the output
    for sample in flow_cell.samples:
        assert sample.internal_id in result.output


def test_get_flow_cell_no_samples_without_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output shows no samples of the flow cell."""
    # GIVEN a database with a flow cell without related samples
    flow_cell: Flowcell = helpers.add_flow_cell(disk_store)

    # WHEN getting a flow cell with the --no-samples flag
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name, "--no-samples"], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS


def test_get_flow_cell_no_samples_with_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flow cell with related samples
    samples: Sample = helpers.add_samples(store=disk_store)
    flow_cell: Flowcell = helpers.add_flow_cell(store=disk_store, samples=samples)

    # WHEN getting a flow cell with the --no-samples flag
    result = cli_runner.invoke(get, ["flow-cell", flow_cell.name, "--no-samples"], obj=base_context)

    # THEN exist successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN no related samples should be listed in the output
    for sample in flow_cell.samples:
        assert sample.internal_id not in result.output



"""Test fixtures for cg/server tests"""

import os
from datetime import datetime
from typing import Generator

import pytest
from flask import Flask
from flask.testing import FlaskClient
from mock import patch

from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import DataDelivery, Workflow
from cg.server.ext import db as store
from cg.store.database import create_all_tables, drop_all_tables
from cg.store.models import Case, Customer, Order, Sample
from tests.store_helpers import StoreHelpers

os.environ["CG_SQL_DATABASE_URI"] = "sqlite:///"
os.environ["LIMS_HOST"] = "dummy_value"
os.environ["LIMS_USERNAME"] = "dummy_value"
os.environ["LIMS_PASSWORD"] = "dummy_value"
os.environ["GOOGLE_OAUTH_CLIENT_ID"] = "dummy_value"
os.environ["GOOGLE_OAUTH_CLIENT_SECRET"] = "dummy_value"
os.environ["CG_ENABLE_ADMIN"] = "1"
os.environ["TRAILBLAZER_SERVICE_ACCOUNT"] = "dummy_value"
os.environ["TRAILBLAZER_SERVICE_ACCOUNT_AUTH_FILE"] = "dummy_value"
os.environ["TRAILBLAZER_HOST"] = "dummy_value"


@pytest.fixture
def app() -> Generator[Flask, None, None]:
    from cg.server.auto import app

    app.config.update({"TESTING": True})
    create_all_tables()
    yield app
    drop_all_tables()


@pytest.fixture
def order(
    helpers: StoreHelpers, customer: Customer, server_case: Case, server_case_in_same_order: Case
) -> Order:
    order: Order = helpers.add_order(
        store=store,
        customer_id=customer.id,
        ticket_id=1,
        order_date=datetime.now(),
        workflow=Workflow.MIP_DNA,
    )
    order.cases.append(server_case)
    order.cases.append(server_case_in_same_order)
    return order


@pytest.fixture
def order_another(helpers: StoreHelpers, customer_another: Customer) -> Order:
    order: Order = helpers.add_order(
        store=store, customer_id=customer_another.id, ticket_id=2, order_date=datetime.now()
    )
    return order


@pytest.fixture
def order_balsamic(helpers: StoreHelpers, customer_another: Customer) -> Order:
    order: Order = helpers.add_order(
        store=store,
        customer_id=customer_another.id,
        ticket_id=3,
        order_date=datetime.now(),
        workflow=Workflow.BALSAMIC,
    )
    return order


@pytest.fixture
def server_samples(helpers: StoreHelpers) -> list[Sample]:
    return helpers.add_samples(store=store, nr_samples=2)


@pytest.fixture
def server_samples_for_another_case(helpers: StoreHelpers) -> list[Sample]:
    return helpers.add_samples(store=store, nr_samples=2)


@pytest.fixture
def server_case(helpers: StoreHelpers, server_samples: list[Sample]) -> Case:
    case: Case = helpers.add_case(
        customer_id=1,
        data_analysis=Workflow.MIP_DNA,
        internal_id="looseygoosey",
        data_delivery=DataDelivery.ANALYSIS_SCOUT,
        name="test case",
        ticket="123",
        store=store,
    )
    helpers.relate_samples(base_store=store, case=case, samples=server_samples)
    return case


@pytest.fixture
def server_case_in_same_order(
    helpers: StoreHelpers, server_samples_for_another_case: list[Sample]
) -> Case:
    case: Case = helpers.add_case(
        customer_id=1,
        data_analysis=Workflow.MIP_DNA,
        data_delivery=DataDelivery.ANALYSIS_SCOUT,
        internal_id="failedsnail",
        name="test case 2",
        ticket="123",
        store=store,
    )
    helpers.relate_samples(base_store=store, case=case, samples=server_samples_for_another_case)
    return case


@pytest.fixture
def trailblazer_analysis_for_server_case(server_case: Case):
    return TrailblazerAnalysis(
        case_id=server_case.internal_id,
        id=1,
        logged_at="",
        started_at="",
        completed_at="",
        out_dir="",
        config_path="",
        uploaded_at="2024-01-01",
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture
def trailblazer_analysis_for_server_case_in_same_order(server_case_in_same_order: Case):
    return TrailblazerAnalysis(
        case_id=server_case_in_same_order.internal_id,
        id=2,
        logged_at="",
        started_at="",
        completed_at="",
        out_dir="",
        config_path="",
        uploaded_at="2023-01-01",
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture
def customer(helpers: StoreHelpers) -> Customer:
    customer: Customer = helpers.ensure_customer(store=store, customer_id="test_customer")
    return customer


@pytest.fixture
def customer_another(helpers: StoreHelpers) -> Customer:
    customer: Customer = helpers.ensure_customer(store=store, customer_id="test_customer_2")
    return customer


@pytest.fixture
def non_existent_order_id() -> int:
    return 900


@pytest.fixture
def client(app: Flask) -> Generator[FlaskClient, None, None]:
    # Bypass authentication
    with patch.object(app, "before_request_funcs", new={}):
        yield app.test_client()


@pytest.fixture
def analysis_summary():
    return AnalysisSummary(
        order_id=1,
        cancelled=0,
        completed=1,
        running=0,
        delivered=1,
        failed=1,
    )



"""Delivery API for handling file delivery to customer's inbox."""

import logging
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import DataDelivery, Workflow
from cg.constants.delivery import INBOX_NAME, PIPELINE_ANALYSIS_TAG_MAP
from cg.models.delivery.delivery import DeliveryFile
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.quality_controller.quality_controller_service import QualityControllerService
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class DeliveryAPI:
    """
    This class provides methods to handle the delivery process of files from the Housekeeper
    bundle to the customer's inbox.
    """

    def __init__(
        self,
        delivery_path: Path,
        fastq_concatenation_service: FastqConcatenationService,
        housekeeper_api: HousekeeperAPI,
        store: Store,
    ):
        self.delivery_path: Path = delivery_path
        self.fastq_file_service: FastqConcatenationService = fastq_concatenation_service
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.store: Store = store

    @staticmethod
    def get_analysis_case_tags_for_workflow(workflow: Workflow) -> list[set[str]]:
        return PIPELINE_ANALYSIS_TAG_MAP[workflow]["case_tags"]

    @staticmethod
    def get_analysis_sample_tags_for_workflow(workflow: Workflow) -> list[set[str]]:
        return PIPELINE_ANALYSIS_TAG_MAP[workflow]["sample_tags"]

    @staticmethod
    def is_fastq_delivery(data_delivery: DataDelivery) -> bool:
        """Return whether FASTQs should be delivered."""
        return DataDelivery.FASTQ in data_delivery

    @staticmethod
    def is_analysis_delivery(data_delivery: DataDelivery) -> bool:
        """Return whether analysis files should be delivered."""
        return DataDelivery.ANALYSIS_FILES in data_delivery

    @staticmethod
    def is_sample_deliverable(sample: Sample, force: bool = False) -> bool:
        """
        Return whether the sample is deliverable or not. A sample is deliverable if it passes QC
        or is external. The force parameter can be used to override checks.
        """
        is_external: bool = sample.application_version.application.is_external
        sample_passed_sequencing_qc: bool = QualityControllerService.sample_pass_sequencing_qc(
            sample=sample
        )

        return is_external or sample_passed_sequencing_qc or force

    def convert_files_to_delivery_files(
        self,
        files: list[File],
        case: Case,
        internal_id: str,
        external_id: str,
        analysis_sample_files: bool = False,
    ) -> list[DeliveryFile]:
        """Return a populated delivery file object given a list of Housekeeper files."""
        delivery_files: list[DeliveryFile] = []
        destination_path = Path(
            self.delivery_path, case.customer.internal_id, INBOX_NAME, case.latest_ticket
        )
        subfolder = Path(external_id)
        if analysis_sample_files:
            subfolder = Path(case.name, external_id)
        for file in files:
            destination_file_name: str = (
                Path(file.full_path)
                .name.replace(internal_id, external_id)
                .replace(case.internal_id, external_id)
            )
            destination_path = Path(destination_path, subfolder, destination_file_name)
            delivery_file = DeliveryFile(
                source_path=Path(file.full_path), destination_path=destination_path
            )
            delivery_files.append(delivery_file)
        return delivery_files

    def get_analysis_sample_delivery_files_by_sample(
        self, case: Case, sample: Sample
    ) -> list[DeliveryFile]:
        """Return a list of analysis files to be delivered for a specific sample."""
        sample_tags: list[set[str]] = self.get_analysis_sample_tags_for_workflow(case.data_analysis)
        sample_tags_with_sample_id: list[set[str]] = [
            tag | {sample.internal_id} for tag in sample_tags
        ]
        sample_files: list[File] = (
            self.housekeeper_api.get_files_from_latest_version_containing_tags(
                bundle_name=case.internal_id, tags=sample_tags_with_sample_id
            )
        )
        delivery_files: list[DeliveryFile] = self.convert_files_to_delivery_files(
            files=sample_files,
            case=case,
            internal_id=sample.internal_id,
            external_id=sample.name,
            analysis_sample_files=True,
        )
        return delivery_files

    def get_analysis_sample_delivery_files(self, case: Case) -> list[DeliveryFile]:
        """Return a complete list of analysis sample files to be delivered."""
        delivery_files: list[DeliveryFile] = []
        for sample in case.samples:
            sample_delivery_files: list[DeliveryFile] = (
                self.get_analysis_sample_delivery_files_by_sample(case=case, sample=sample)
            )
            delivery_files.extend(sample_delivery_files)
        return delivery_files

    def get_analysis_case_delivery_files(self, case: Case) -> list[DeliveryFile]:
        """
        Return a complete list of analysis case files to be delivered and ignore analysis sample
        files.
        """
        case_tags: list[set[str]] = self.get_analysis_case_tags_for_workflow(case.data_analysis)
        sample_id_tags: list[str] = [sample_id for sample_id in case.sample_ids]
        case_files: list[File] = self.housekeeper_api.get_files_from_latest_version_containing_tags(
            bundle_name=case.internal_id, tags=case_tags, excluded_tags=sample_id_tags
        )
        delivery_files: list[DeliveryFile] = self.convert_files_to_delivery_files(
            files=case_files, case=case, internal_id=case.internal_id, external_id=case.name
        )
        return delivery_files

    def get_fastq_delivery_files_by_sample(
        self, case: Case, sample: Sample, force: bool = False
    ) -> list[DeliveryFile]:
        """Return a list of FASTQ files to be delivered for a specific sample."""
        delivery_files: list[DeliveryFile] = []
        fastq_tags: list[set[str]] = self.get_analysis_sample_tags_for_workflow(Workflow.FASTQ)
        if not self.is_sample_deliverable(sample=sample, force=force):
            LOG.warning(f"Sample {sample.internal_id} is not deliverable")
            return delivery_files
        fastq_files: list[File] = (
            self.housekeeper_api.get_files_from_latest_version_containing_tags(
                bundle_name=sample.internal_id, tags=fastq_tags
            )
        )
        delivery_files: list[DeliveryFile] = self.convert_files_to_delivery_files(
            files=fastq_files,
            case=case,
            internal_id=sample.internal_id,
            external_id=sample.name,
        )
        return delivery_files

    def get_fastq_delivery_files(self, case: Case, force: bool = False) -> list[DeliveryFile]:
        """Return a complete list of FASTQ sample files to be delivered."""
        delivery_files: list[DeliveryFile] = []
        for sample in case.samples:
            fastq_delivery_files: list[DeliveryFile] = self.get_fastq_delivery_files_by_sample(
                case=case, sample=sample, force=force
            )
            delivery_files.extend(fastq_delivery_files)
        return delivery_files

    def get_delivery_files(self, case: Case, force: bool = False) -> list[DeliveryFile]:
        """Return a complete list of files to be delivered to the customer's inbox."""
        delivery_files: list[DeliveryFile] = []
        if self.is_fastq_delivery(case.data_delivery):
            fastq_files: list[DeliveryFile] = self.get_fastq_delivery_files(case=case, force=force)
            delivery_files.extend(fastq_files)
        if self.is_analysis_delivery(case.data_delivery):
            analysis_case_files: list[DeliveryFile] = self.get_analysis_case_delivery_files(case)
            analysis_sample_files: list[DeliveryFile] = self.get_analysis_sample_delivery_files(
                case=case
            )
            delivery_files.extend(analysis_case_files + analysis_sample_files)
        return delivery_files

    def link_delivery_files_to_inbox(self) -> None:
        """Link files from Housekeeper bundle to the customer's inbox."""
        raise NotImplementedError
